-- USDP Sample code
-- Access control system
-- Active 'LOW' reset
-- User has to enter passowrd "143" on keypad, then press F1 twice to validate the password.
-- The alarm will go high if password if wrong, then user has to reset the system.
-- The relay o/p will go high is password is ok, then user has to press any other key
--	 to close relay and get back in system.

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity access_control is
generic (low: std_logic:= '0');
    Port ( 
			clk			: in  std_logic; 	    -- clock
      	Rst         : in  std_logic; 	    -- System Reset       
        	RL          : in  std_logic_vector(3 downto 0);		-- return lines
       	SL          : out std_logic_vector(3 downto 0);		-- scan lines
        	hit         : out std_logic;						-- key hit signal
			rst_51		: out std_logic;						-- 89c51 reset
			ready			: out std_logic;						-- system ready indication
			ready_51		: out std_logic;						-- system ready indication to 89c51
			relay			: out std_logic;						-- relay control
			ok_led		: out std_logic;						-- ok led status
			Alarm			: out std_logic;						-- Alarm o/p
			Pass_OK		: out std_logic;						-- passowrd ok
			Pass_Err		: out std_logic						-- password mis-matched
			);
end access_control;

architecture Behavioral of access_control is
signal div 		     : std_logic_vector(14 downto 0);
--- keypad signal declarations
signal SL_s          :  std_logic_vector(3 downto 0);
signal D		     :  std_logic_vector(5 downto 0);
signal key_code_s    :  std_logic_vector(7 downto 0);
signal key_value_s    :  std_logic_vector(4 downto 0);
signal Scan_Clk	: std_logic; 	--scan clock
signal Deb_Clk	: std_logic;  	-- debounce clock          	    
signal or_key        :  std_logic;
signal hit_s,deb_in  :  std_logic;
signal stop          :  std_logic;
signal en_s 	     	: std_logic;
signal ready_s      	: std_logic;
signal release       :  std_logic;
--- FSM signal declarations
type states is(reset, key1,key2,key3,ok,done,false1,false2,false3,error);
signal ps,nx	: states;	-- present and next state declaration
signal alarm_s      :  std_logic;
signal relay_s      :  std_logic;
begin
rst_51	<= not rst;
ok_led	<= relay_s;
Scan_Clk  <= div(14) when stop='0' else
				 '0';
Deb_Clk  <= div(8);
-- key hit indication
hit 		<= hit_s;
SL  		<= sl_s;
relay  	<= relay_s;
alarm		<= alarm_s;
pass_err	<= alarm_s;		-- password mismatched
pass_OK	<= relay_s;		-- password OK indication to controller.
ready		<= ready_s;
ready_51	<= ready_s;

-- Clock divider
    process(clk,rst)
    begin
    	if rst = low then 
    		div	<= (others=>'0');
    	 elsif clk'event and clk='1' then
        	div <= div +1;
        end if ;
    end process;

----------------- FSM Logic for access control --------------
-- present state memory	 
process(hit_s,rst)
begin
    	if rst = low then 
			ps	<= reset;
		elsif hit_s'event and hit_s='1' then
			ps	<= nx;
		end if;
end process;

-- Next state decoder
process(ps,key_value_s)
begin

case(ps) is
when reset	=> nx	<= key1;

when key1	=> if key_value_s="00001" then	-- 1st key value "1"
						nx	<= key2;
					else
						nx	<= false1;
					end if;

when key2	=> if key_value_s="00100" then	-- 2nd key value "4"
						nx	<= key3;
					else
						nx	<= false2;
					end if;

when key3	=> if key_value_s="00011" then	-- 3rd key value "3"
						nx	<= ok;
					else
						nx	<= false3;
					end if;

when ok	=> if key_value_s="01010" then	-- if "F1" Key pressed
						nx	<= done;
					else
						nx	<= ok;
					end if;
when done	=> nx	<= reset;
when false1	=> nx	<= false2;
when false2	=> nx	<= false3;
when false3	=> nx	<= error;
when error	=> nx	<= error;
when others => nx	<= reset;
end case;
end process;

-- Assignning O/Ps
process(ps)
begin
case(ps) is
when reset	=> ready_s		<= '1';
					alarm_s	<= '0';
					relay_s	<= '0';
when done	=> ready_s		<= '0';
					alarm_s	<= '0';
					relay_s	<= '1';
when error	=> ready_s		<= '0';
					alarm_s	<= '1';
					relay_s	<= '0';
when others => ready_s		<= '0';
					alarm_s	<= '0';
					relay_s	<= '0';
end case;
end process;
---------------------------------------------------------------

-- Keypad controller code
-- stopping the scanning process when key pressed
	process(hit_s,release,Rst)
	begin
		if(rst = low) or release='1' then
			stop	<= '0';
		elsif(hit_s='1') then
			stop	<= '1';
		end if;
	end process;
---------------------------------------------------------------
--	SCAN PATTERN GENERATOR
	process(Scan_Clk,Rst)
	begin
		if(rst = low)then
			sl_s <= "1110";
		elsif(Scan_Clk'event and Scan_Clk = '0')then
    			sl_s <= SL_s(0) & sl_s(3 downto 1);
		end if;
	end process;
---------------------------------------------------------------
-- process for key press detection
deb_in <= '1' when RL(0)='0' or RL(1)='0' or RL(2)='0' or RL(3)='0' else
          '0';
---------------------------------------------------------------
--	DEBOUNCING THE 'OR' VALUE FOR 6 CLOCK CYCLES
	process(Deb_Clk,Rst)
	begin
		if(rst = low)then
			D <= (others => '0');
		elsif(Deb_Clk'event and Deb_Clk = '0')then
			D <= deb_in & D(5 downto 1);
	    end if;
	end process;
---------------------------------------------------------------
-- GENERATING INTERRUPT IF FOR 6 CLOCKS THE KEY IS PRESSED
-- Valid key hit
hit_s <= D(0) and D(1) and D(2) and D(3) and D(4) and D(5);

-- key released
release	<= not hit_s;      
---------------------------------------------------------------
--	FOR LATCHING THE VALID KEY PRESSED
	process(hit_s,RL,Rst )
	begin
		if rst = low then
			key_code_s <= (others=>'0');
		elsif( hit_s = '1'and hit_s'event )then
			key_code_s <= sl_s & RL;
        end if;
	end process;
-- binary coding for keys
-- MSB is for false key value detected.
with key_code_s select
key_value_s <=	"00000" when "11101110",   -- 0
           		"00001" when "11101101",   -- 1    
           		"00010" when "11101011",   -- 2
            	"00011" when "11100111",   -- 3
	            "00100" when "11011110",   -- 4
	            "00101" when "11011101",   -- 5
	            "00110" when "11011011",   -- 6
	            "00111" when "11010111",   -- 7
	            "01000" when "10111110",   -- 8
	            "01001" when "10111101",   -- 9
	            "01010" when "10111011",   -- F1
	            "01011" when "10110111",   -- F2
	            "01100" when "01111110",   -- F3
	            "01101" when "01111101",   -- F4
	            "01110" when "01111011",   -- F5
	            "01111" when "01110111",   -- F6
	            "10000" when others;
---------------------------------------------------------------
end Behavioral;
